@@ -46,7 +46,8 @@
 #define MIME_READ_CHUNK_SIZE 1024
 
 typedef gboolean (*PlaylistCallback) (TotemPlParser *parser, const char *url, gpointer data);
-static gboolean totem_pl_parser_ignore (const char *url);
+static gboolean totem_pl_parser_scheme_is_ignored (TotemPlParser *parser, const char *url);
+static gboolean totem_pl_parser_ignore (TotemPlParser *parser, const char *url);
 
 typedef struct {
 	char *mimetype;
@@ -55,9 +56,7 @@ typedef struct {
 
 struct TotemPlParserPrivate
 {
-	GladeXML *xml;
-
-	int x, y;
+	GList *ignore_schemes;
 };
 
 /* Signals */
@@ -216,17 +215,48 @@ write_string (GnomeVFSHandle *handle, const char *buf, GError **error)
 	return TRUE;
 }
 
+static int
+totem_pl_parser_num_entries (TotemPlParser *parser, GtkTreeModel *model,
+			     TotemPlParserIterFunc func)
+{
+	int num_entries, i, ignored;
+
+	num_entries = gtk_tree_model_iter_n_children (model, NULL);
+	ignored = 0;
+
+	for (i = 1; i <= num_entries; i++)
+	{
+		GtkTreeIter iter;
+		char *path, *url, *title;
+
+		path = g_strdup_printf ("%d", i - 1);
+		gtk_tree_model_get_iter_from_string (model, &iter, path);
+		g_free (path);
+
+		func (model, &iter, &url, &title);
+		if (totem_pl_parser_scheme_is_ignored (parser, url) != FALSE)
+			ignored++;
+
+		g_free (url);
+		g_free (title);
+	}
+
+	return num_entries - ignored;
+}
+
 gboolean
 totem_pl_parser_write (TotemPlParser *parser, GtkTreeModel *model,
 		   TotemPlParserIterFunc func, const char *output, GError **error)
 {
 	GnomeVFSHandle *handle;
 	GnomeVFSResult res;
-	int num_entries, i;
+	int num_entries_total, num_entries, i;
 	char *buf;
 	gboolean success;
 
-	num_entries = gtk_tree_model_iter_n_children (model, NULL);
+	num_entries = totem_pl_parser_num_entries (parser, model, func);
+	num_entries_total = gtk_tree_model_iter_n_children (model, NULL);
+
 	res = gnome_vfs_open (&handle, output, GNOME_VFS_OPEN_WRITE);
 	if (res == GNOME_VFS_ERROR_NOT_FOUND) {
 		res = gnome_vfs_create (&handle, output,
@@ -251,23 +281,30 @@ totem_pl_parser_write (TotemPlParser *parser, GtkTreeModel *model,
 	if (success == FALSE)
 		return FALSE;
 
-	buf = g_strdup_printf ("numberofentries=%d\n", num_entries);
+	buf = g_strdup_printf ("NumberOfEntries=%d\n", num_entries);
 	success = write_string (handle, buf, error);
 	g_free (buf);
 	if (success == FALSE)
 		return FALSE;
 
-	for (i = 1; i <= num_entries; i++) {
+	for (i = 1; i <= num_entries_total; i++) {
 		GtkTreeIter iter;
 		char *path, *url, *title;
 
 		path = g_strdup_printf ("%d", i - 1);
 		gtk_tree_model_get_iter_from_string (model, &iter, path);
 		g_free (path);
-		
+
 		func (model, &iter, &url, &title);
 
-		buf = g_strdup_printf ("file%d=%s\n", i, url);
+		if (totem_pl_parser_scheme_is_ignored (parser, url) != FALSE)
+		{
+			g_free (url);
+			g_free (title);
+			continue;
+		}
+
+		buf = g_strdup_printf ("File%d=%s\n", i, url);
 		success = write_string (handle, buf, error);
 		g_free (buf);
 		g_free (url);
@@ -277,7 +314,7 @@ totem_pl_parser_write (TotemPlParser *parser, GtkTreeModel *model,
 			return FALSE;
 		}
 
-		buf = g_strdup_printf ("title%d=%s\n", i, title);
+		buf = g_strdup_printf ("Title%d=%s\n", i, title);
 		success = write_string (handle, buf, error);
 		g_free (buf);
 		g_free (title);
@@ -440,6 +477,8 @@ totem_pl_parser_finalize (GObject *object)
 	g_return_if_fail (object != NULL);
 	g_return_if_fail (parser->priv != NULL);
 
+	g_list_foreach (parser->priv->ignore_schemes, (GFunc) g_free, NULL);
+
 	if (G_OBJECT_CLASS (parent_class)->finalize != NULL) {
 		(* G_OBJECT_CLASS (parent_class)->finalize) (object);
 	}
@@ -462,12 +501,12 @@ totem_pl_parser_add_one_url_ext (TotemPlParser *parser, const char *url, const c
 }
 
 static gboolean
-totem_pl_parser_add_m3u (TotemPlParser *parser, const char *url, gpointer data)
+totem_pl_parser_add_ram (TotemPlParser *parser, const char *url, gpointer data)
 {
 	gboolean retval = FALSE;
 	char *contents, **lines;
 	int size, i;
-	char *split_char;
+	const char *split_char;
 
 	if (my_eel_read_entire_file (url, &size, &contents) != GNOME_VFS_OK)
 		return FALSE;
@@ -475,7 +514,7 @@ totem_pl_parser_add_m3u (TotemPlParser *parser, const char *url, gpointer data)
 	contents = g_realloc (contents, size + 1);
 	contents[size] = '\0';
 
-	/* figure out whether we're a unix m3u or dos m3u */
+	/* figure out whether we're a unix or dos RAM file */
 	if (strstr(contents,"\x0d") == NULL)
 		split_char = "\n";
 	else
@@ -493,13 +532,103 @@ totem_pl_parser_add_m3u (TotemPlParser *parser, const char *url, gpointer data)
 		/* Either it's a URI, or it has a proper path ... */
 		if (strstr(lines[i], "://") != NULL
 				|| lines[i][0] == G_DIR_SEPARATOR) {
-			/* We use the same code for .ram and m3u parsers,
-			 * and .ram files can contain .smil entries */
+			/* .ram files can contain .smil entries */
 			if (totem_pl_parser_parse (parser, lines[i]) == FALSE)
 			{
 				totem_pl_parser_add_one_url (parser,
 						lines[i], NULL);
 			}
+		} else if (strcmp (lines[i], "--stop--") == 0) {
+			/* For Real Media playlists, handle the stop command */
+			break;
+		} else {
+			char *fullpath, *base;
+
+			/* Try with a base */
+			base = totem_pl_parser_base_url (url);
+
+			fullpath = g_strdup_printf ("%s/%s", base, lines[i]);
+			if (totem_pl_parser_parse (parser, fullpath) == FALSE)
+			{
+				totem_pl_parser_add_one_url (parser, fullpath, NULL);
+			}
+			g_free (fullpath);
+			g_free (base);
+		}
+	}
+
+	g_strfreev (lines);
+
+	return retval;
+}
+
+static const char *
+totem_pl_parser_get_extinfo_title (gboolean extinfo, char **lines, int i)
+{
+	const char *retval;
+
+	if (extinfo == FALSE)
+		return NULL;
+
+	if (i == 0)
+		return NULL;
+
+	retval = strstr (lines[i-1], "#EXTINF:");
+	retval = strstr (retval, ",");
+	if (retval == NULL || retval[0] == '\0')
+		return NULL;
+
+	retval++;
+
+	return retval;
+}
+
+static gboolean
+totem_pl_parser_add_m3u (TotemPlParser *parser, const char *url, gpointer data)
+{
+	gboolean retval = FALSE;
+	char *contents, **lines;
+	int size, i;
+	const char *split_char;
+	gboolean extinfo;
+
+	if (my_eel_read_entire_file (url, &size, &contents) != GNOME_VFS_OK)
+		return FALSE;
+
+	contents = g_realloc (contents, size + 1);
+	contents[size] = '\0';
+
+	/* is TRUE if there's an EXTINF on the previous line */
+	extinfo = FALSE;
+
+	/* figure out whether we're a unix m3u or dos m3u */
+	if (strstr(contents,"\x0d") == NULL)
+		split_char = "\n";
+	else
+		split_char = "\x0d\n";
+
+	lines = g_strsplit (contents, split_char, 0);
+	g_free (contents);
+
+	for (i = 0; lines[i] != NULL; i++) {
+		if (lines[i][0] == '\0')
+			continue;
+
+		retval = TRUE;
+
+		/* Ignore comments, but mark it if we have extra info */
+		if (lines[i][0] == '#') {
+			if (strstr (lines[i], "#EXTINF") != NULL)
+				extinfo = TRUE;
+			continue;
+		}
+
+		/* Either it's a URI, or it has a proper path ... */
+		if (strstr(lines[i], "://") != NULL
+				|| lines[i][0] == G_DIR_SEPARATOR) {
+			totem_pl_parser_add_one_url (parser, lines[i],
+					totem_pl_parser_get_extinfo_title (extinfo, lines, i));
+			extinfo = FALSE;
 		} else if (lines[i][0] == '\\' && lines[i][1] == '\\') {
 			/* ... Or it's in the windows smb form
 			 * (\\machine\share\filename), Note drive names
@@ -510,8 +639,9 @@ totem_pl_parser_add_m3u (TotemPlParser *parser, const char *url, gpointer data)
 			lines[i] = g_strdelimit (lines[i], "\\", '/');
 			tmpurl = g_strjoin (NULL, "smb:", lines[i], NULL);
 
-			totem_pl_parser_add_one_url (parser, lines[i], NULL);
-			retval = TRUE;
+			totem_pl_parser_add_one_url (parser, lines[i],
+					totem_pl_parser_get_extinfo_title (extinfo, lines, i));
+			extinfo = FALSE;
 
 			g_free (tmpurl);
 		} else {
@@ -523,11 +653,8 @@ totem_pl_parser_add_m3u (TotemPlParser *parser, const char *url, gpointer data)
 			if (sep == '\\')
 				lines[i] = g_strdelimit (lines[i], "\\", '/');
 			fullpath = g_strdup_printf ("%s/%s", base, lines[i]);
-			if (totem_pl_parser_parse (parser, fullpath) == FALSE)
-			{
-				totem_pl_parser_add_one_url (parser,
-						fullpath, NULL);
-			}
+			totem_pl_parser_add_one_url (parser, fullpath,
+					totem_pl_parser_get_extinfo_title (extinfo, lines, i));
 			g_free (fullpath);
 			g_free (base);
 		}
@@ -586,6 +713,12 @@ totem_pl_parser_add_pls (TotemPlParser *parser, const char *url, gpointer data)
 	if (my_eel_read_entire_file (url, &size, &contents) != GNOME_VFS_OK)
 		return FALSE;
 
+	if (size == 0)
+	{
+		g_free (contents);
+		return TRUE;
+	}
+
 	contents = g_realloc (contents, size + 1);
 	contents[size] = '\0';
 
@@ -708,7 +841,7 @@ parse_asx_entries (TotemPlParser *parser, char *base, xmlDocPtr doc,
 
 		if (g_ascii_strcasecmp (node->name, "entry") == 0) {
 			/* Whee found an entry here, find the REF and TITLE */
-			if (parse_asx_entry (parser, base, doc, node) == TRUE)
+			if (parse_asx_entry (parser, base, doc, node) != FALSE)
 				retval = TRUE;
 		}
 	}
@@ -746,7 +879,7 @@ totem_pl_parser_add_asx (TotemPlParser *parser, const char *url, gpointer data)
 	base = totem_pl_parser_base_url (url);
 
 	for (node = doc->children; node != NULL; node = node->next)
-		if (parse_asx_entries (parser, base, doc, node) == TRUE)
+		if (parse_asx_entries (parser, base, doc, node) != FALSE)
 			retval = TRUE;
 
 	g_free (base);
@@ -765,8 +898,7 @@ totem_pl_parser_add_ra (TotemPlParser *parser, const char *url, gpointer data)
 		return TRUE;
 	}
 
-	/* How nice, same format as m3u it seems */
-	return totem_pl_parser_add_m3u (parser, url, NULL);
+	return totem_pl_parser_add_ram (parser, url, NULL);
 }
 
 static gboolean
@@ -810,7 +942,7 @@ parse_smil_entry (TotemPlParser *parser, char *base, xmlDocPtr doc,
 
 			if (url != NULL) {
 				if (parse_smil_video_entry (parser,
-						base, url, title) == TRUE)
+						base, url, title) != FALSE)
 					retval = TRUE;
 			}
 
@@ -818,7 +950,7 @@ parse_smil_entry (TotemPlParser *parser, char *base, xmlDocPtr doc,
 			g_free (url);
 		} else {
 			if (parse_smil_entry (parser,
-						base, doc, node) == TRUE)
+						base, doc, node) != FALSE)
 				retval = TRUE;
 		}
 	}
@@ -839,7 +971,7 @@ parse_smil_entries (TotemPlParser *parser, char *base, xmlDocPtr doc,
 
 		if (g_ascii_strcasecmp (node->name, "body") == 0) {
 			if (parse_smil_entry (parser, base,
-						doc, node) == TRUE)
+						doc, node) != FALSE)
 				retval = TRUE;
 		}
 
@@ -881,7 +1013,7 @@ totem_pl_parser_add_smil (TotemPlParser *parser, const char *url, gpointer data)
 	base = totem_pl_parser_base_url (url);
 
 	for (node = doc->children; node != NULL; node = node->next)
-		if (parse_smil_entries (parser, base, doc, node) == TRUE)
+		if (parse_smil_entries (parser, base, doc, node) != FALSE)
 			retval = TRUE;
 
 	return FALSE;
@@ -931,7 +1063,7 @@ totem_pl_parser_add_desktop (TotemPlParser *parser, const char *url, gpointer da
 	retval = TRUE;
 
 	display_name = gnome_desktop_item_get_localestring (ditem, "Name");
-	if (totem_pl_parser_ignore (path) == FALSE) {
+	if (totem_pl_parser_ignore (parser, path) == FALSE) {
 		totem_pl_parser_add_one_url (parser, path, display_name);
 	} else {
 		if (totem_pl_parser_parse (parser, path) == FALSE)
@@ -967,8 +1099,7 @@ totem_pl_parser_add_directory (TotemPlParser *parser, const char *url,
 			continue;
 		}
 
-		str = g_build_filename (G_DIR_SEPARATOR_S,
-					     url, info->name, NULL);
+		str = g_build_filename (url, info->name, NULL);
 		if (strstr (str, "://") != NULL)
 			fullpath = str + 1;
 		else
@@ -997,7 +1128,8 @@ static PlaylistTypes special_types[] = {
 	{ "application/x-gnome-app-info", totem_pl_parser_add_desktop },
 #endif	
 	{ "x-directory/normal", totem_pl_parser_add_directory },
-	{ "video/x-ms-wvx", totem_pl_parser_add_asf_parser },
+	{ "video/x-ms-wvx", totem_pl_parser_add_asx },
+	{ "audio/x-ms-wax", totem_pl_parser_add_asx },
 };
 
 /* These ones are "dual" types, might be a video, might be a parser */
@@ -1011,7 +1143,23 @@ static PlaylistTypes dual_types[] = {
 	{ "video/x-ms-wmv", totem_pl_parser_add_asf },
 };
 
-static char * ignore_mime[] = { "cdda:", "dvd:", "vcd:" };
+static gboolean
+totem_pl_parser_scheme_is_ignored (TotemPlParser *parser, const char *url)
+{
+	GList *l;
+
+	if (parser->priv->ignore_schemes == NULL)
+		return FALSE;
+
+	for (l = parser->priv->ignore_schemes; l != NULL; l = l->next)
+	{
+		const char *scheme = l->data;
+		if (strncmp (url, scheme, strlen (scheme)) == 0)
+			return TRUE;
+	}
+
+	return FALSE;
+}
 
 static gboolean
 totem_pl_parser_add_url_from_data (TotemPlParser *parser, const char *url)
@@ -1049,14 +1197,13 @@ totem_pl_parser_add_url_from_data (TotemPlParser *parser, const char *url)
 }
 
 static gboolean
-totem_pl_parser_ignore (const char *url)
+totem_pl_parser_ignore (TotemPlParser *parser, const char *url)
 {
 	const char *mimetype;
 	guint i;
 
-	for (i = 0; i < G_N_ELEMENTS(ignore_mime); i++)
-		if (strncmp (url, ignore_mime[i], strlen (ignore_mime[i])) == 0)
-			return TRUE;
+	if (totem_pl_parser_scheme_is_ignored (parser, url) != FALSE)
+		return TRUE;
 
 	mimetype = gnome_vfs_get_file_mime_type (url, NULL, TRUE);
 	if (mimetype == NULL || strcmp (mimetype, "application/octet-stream") == 0)
@@ -1079,9 +1226,10 @@ totem_pl_parser_parse (TotemPlParser *parser, const char *url)
 	const char *mimetype;
 	guint i;
 
+	g_return_val_if_fail (TOTEM_IS_PL_PARSER (parser), FALSE);
 	g_return_val_if_fail (url != NULL, FALSE);
 
-	if (totem_pl_parser_ignore (url) != FALSE)
+	if (totem_pl_parser_ignore (parser, url) != FALSE)
 		return FALSE;
 
 	mimetype = gnome_vfs_get_mime_type (url);
@@ -1101,3 +1249,14 @@ totem_pl_parser_parse (TotemPlParser *parser, const char *url)
 	return TRUE;
 }
 
+void
+totem_pl_parser_add_ignored_scheme (TotemPlParser *parser,
+		const char *scheme)
+{
+	g_return_if_fail (TOTEM_IS_PL_PARSER (parser));
+
+	parser->priv->ignore_schemes = g_list_prepend
+		(parser->priv->ignore_schemes, g_strdup (scheme));
+}
+
+
